/* Copyright 2013-2019 Homegear GmbH
 *
 * libhomegear-base is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * libhomegear-base is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with libhomegear-base.  If not, see
 * <http://www.gnu.org/licenses/>.
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU Lesser General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
*/

#include "Scenario.h"
#include "../BaseLib.h"

namespace BaseLib
{
namespace DeviceDescription
{

Scenario::Scenario(BaseLib::SharedObjects* baseLib)
{
	_bl = baseLib;
}

Scenario::Scenario(BaseLib::SharedObjects* baseLib, xml_node<>* node) : Scenario(baseLib)
{
	for(xml_attribute<>* attr = node->first_attribute(); attr; attr = attr->next_attribute())
	{
		std::string attributeName(attr->name());
		std::string attributeValue(attr->value());
		if(attributeName == "id")
		{
			id = attributeValue;
		}
		else _bl->out.printWarning("Warning: Unknown attribute for \"scenario\": " + attributeName);
	}
	for(xml_node<>* subNode = node->first_node(); subNode; subNode = subNode->next_sibling())
	{
		std::string nodeName(subNode->name());
		std::string parameterId;
		if(nodeName == "parameter")
		{
			for(xml_attribute<>* attr = subNode->first_attribute(); attr; attr = attr->next_attribute())
			{
				std::string attributeName(attr->name());
				if(attributeName == "id") parameterId = std::string(attr->value());
				else _bl->out.printWarning("Warning: Unknown attribute for \"scenario\\parameter\": " + std::string(attr->name()));
			}
			if(parameterId.empty()) continue;
			scenarioEntries[parameterId] = std::string(subNode->value());
		}
		else _bl->out.printWarning("Warning: Unknown node in \"scenario\": " + nodeName);
	}
}

Scenario::~Scenario()
{
}

}
}
